"""Deprecated: Use `guppylang.std.quantum.functional` instead."""

from guppylang.std.quantum.functional import (
    ch,
    crz,
    cx,
    cy,
    cz,
    h,
    project_z,
    reset,
    rx,
    ry,
    rz,
    s,
    sdg,
    t,
    tdg,
    toffoli,
    v,
    vdg,
    x,
    y,
    z,
)

__all__ = (
    "h",
    "cz",
    "cx",
    "cy",
    "t",
    "s",
    "v",
    "x",
    "y",
    "z",
    "tdg",
    "sdg",
    "vdg",
    "rz",
    "rx",
    "ry",
    "crz",
    "toffoli",
    "reset",
    "project_z",
    "ch",
)
